import ast

from collections import defaultdict
from bs4 import BeautifulSoup

IGNORED_AXTREE_ROLES = ["LineBreak"]

IGNORED_AXTREE_PROPERTIES = (
    "editable",
    "readonly",
    "level",
    "settable",
    "multiline",
    "invalid",
    "focusable",
)


def flatten_dom_to_str(
    DOM_tree,
    with_visible: bool = False,
    with_center_coords: bool = False,
    with_bounding_box_coords: bool = False,
    filter_visible_only: bool = False,
    coord_decimals: int = 0,
) -> str:
    """Formats a DOM snapshot into a string text"""

    coord_format = f":0.{coord_decimals}f"

    def parse_DOM(document_idx) -> str:
        # adopted from [natbot](https://github.com/nat/natbot)

        strings = DOM_tree["strings"]
        nodes = DOM_tree["documents"][document_idx]["nodes"]
        node_iframe_link = nodes["contentDocumentIndex"]
        graph = defaultdict(lambda: [])

        for node_idx in range(len(nodes["nodeName"])):
            parent_idx = nodes["parentIndex"][node_idx]
            if parent_idx != -1:
                graph[parent_idx].append(node_idx)

        def dfs(idx: int) -> str:
            node_name = strings[nodes["nodeName"][idx]]
            can_skip = (
                "#" in node_name or "::" in node_name or node_name == "html"
            )  # We skip the root <DOCTYPE html> node
            node_name = node_name.lower().strip()
            html = ""

            # print node opening tag
            if not can_skip:
                html += f"<{node_name}"
                node_attr_idxs = nodes["attributes"][idx]
                if node_attr_idxs:
                    for i in range(0, len(node_attr_idxs), 2):
                        attr_name = strings[node_attr_idxs[i]]

                        # filter visible elements if requested
                        if (
                            filter_visible_only
                            and attr_name == "browsergym_is_in_viewport"
                            and strings[node_attr_idxs[i + 1]] == "0"
                        ):
                            can_skip = True
                            break

                        # print browsergym attributes if requested (with new names)
                        if attr_name == "browsergym_is_in_viewport":
                            if with_visible:
                                attr_value = strings[node_attr_idxs[i + 1]]
                                html += f' is_visible="{attr_value}"'
                        elif attr_name == "browsergym_center":
                            if with_center_coords:
                                attr_value = strings[node_attr_idxs[i + 1]]
                                html += f' center="{_get_coord_str(attr_value, coord_decimals)}"'

                        elif attr_name == "browsergym_bounding_box":
                            if with_bounding_box_coords:
                                attr_value = strings[node_attr_idxs[i + 1]]
                                html += f' box="{_get_coord_str(attr_value, coord_decimals)}"'

                        # print other attributes
                        else:
                            if node_attr_idxs[i + 1] >= 0:
                                attr_value = strings[node_attr_idxs[i + 1]]
                                # attribute value present
                                html += f' {attr_name}="{attr_value}"'
                            else:
                                # attribute value missing
                                html += f" {attr_name}"

                html += f">"
            if can_skip:
                html = ""
            # print inner text
            node_value_idx = nodes["nodeValue"][idx]
            if node_value_idx >= 0:
                html += " ".join(strings[node_value_idx].split())

            # recursively print iframe nodes if any
            if idx in node_iframe_link["index"]:
                sub_document_idx = node_iframe_link["value"][node_iframe_link["index"].index(idx)]
                html += parse_DOM(document_idx=sub_document_idx)

            # recursively print children nodes
            for child_idx in graph[idx]:
                html += dfs(child_idx)

            # print node closing tag
            if not can_skip:
                html += f"</{node_name}>"

            return html

        html = dfs(idx=0)

        # Format the HTML document with indentation
        soup = BeautifulSoup(html, "lxml")
        html = soup.prettify()

        return html

    html = parse_DOM(0)

    return html


def _remove_redundant_static_text(ax_tree: str) -> str:
    """Removes redundant `StaticText` from the accessibility tree"""
    new_lines = []
    lines = ax_tree.split("\n")
    for line in lines:
        if line.strip().startswith("StaticText"):
            content = line.split("StaticText")[1].strip().strip("'")
            if content in "\n".join(new_lines[-3:]):
                continue
        new_lines.append(line)
    return "\n".join(new_lines)


def _get_coord_str(coord, decimals):
    if isinstance(coord, str):
        coord = list(map(float, ast.literal_eval(coord)))

    coord_format = f".{decimals}f"
    coord_str = ",".join([f"{c:{coord_format}}" for c in coord])
    return f"({coord_str})"


def flatten_axtree_to_str(
    AX_tree,
    with_visible: bool = False,
    with_center_coords: bool = False,
    with_bounding_box_coords: bool = False,
    filter_visible_only: bool = False,
    coord_decimals: int = 0,
    ignore_roles=IGNORED_AXTREE_ROLES,
    ignored_properties=IGNORED_AXTREE_PROPERTIES,
    remove_rdundant_static_text: bool = True,
) -> str:
    """Formats the accessibility tree into a string text"""
    node_id_to_idx = {}
    for idx, node in enumerate(AX_tree["nodes"]):
        node_id_to_idx[node["nodeId"]] = idx

    def dfs(idx: int, depth: int) -> str:
        tree_str = ""
        node = AX_tree["nodes"][idx]
        indent = "\t" * depth
        valid_node = True
        role = node["role"]["value"]

        if role in ignore_roles:
            pass
        elif "name" not in node:
            pass
        else:
            print_node = True
            name = node["name"]["value"]
            node_str = f"{role} {repr(name.strip())}"

            if "value" in node and "value" in node["value"]:
                node_str += f' value: {repr(node["value"]["value"])}'

            properties = []
            for property in node.get("properties", []):
                if not "value" in property:
                    continue
                if not "value" in property["value"]:
                    continue

                prop_name, value = property["name"], property["value"]["value"]
                if prop_name == "browsergym_id":
                    node_str = f"[{value}] " + node_str
                elif prop_name == "browsergym_center":
                    if with_center_coords:
                        coord_str = _get_coord_str(value, coord_decimals)
                        node_str += f", center={coord_str}"
                elif prop_name == "browsergym_bounding_box":
                    if with_bounding_box_coords:
                        coord_str = _get_coord_str(value, coord_decimals)
                        node_str += f", box={coord_str}"
                elif prop_name == "browsergym_is_in_viewport":
                    # filter visible elements if requested
                    if filter_visible_only and not value:
                        print_node = False
                        break
                    if with_visible:
                        visible_str = "visible" if value else "hidden"
                        node_str += f", {visible_str}"
                elif prop_name in ("required", "focused", "atomic"):
                    if value:
                        properties.append(prop_name)
                elif prop_name not in ignored_properties:
                    properties.append(f"{prop_name}={repr(value)}")

            if role in ["generic"] and not properties:
                print_node = False

            if properties:
                node_str += " " + ", ".join(properties)

            if print_node:
                tree_str += f"{indent}{node_str}"

        for _, child_node_id in enumerate(node["childIds"]):
            if child_node_id not in node_id_to_idx or child_node_id == node["nodeId"]:
                continue
            # mark this to save some tokens
            child_depth = depth + 1 if valid_node else depth
            child_str = dfs(node_id_to_idx[child_node_id], child_depth)
            if child_str.strip():
                if tree_str.strip():
                    tree_str += "\n"
                tree_str += child_str

        return tree_str

    tree_str = dfs(0, 0)
    if remove_rdundant_static_text:
        tree_str = _remove_redundant_static_text(tree_str)
    return tree_str



from typing import Type

import gymnasium as gym
from .env import BrowserEnv
from .task import AbstractBrowserTask


def register_task(id: str, task_class: Type[AbstractBrowserTask], *args, **kwargs):
    """
    Registers a browser task as a gym environment with its unique id.

    Args:
        task: the task class to register.
    """

    gym.register(
        id=f"browsergym/{id}",
        entry_point=lambda *args, **kwargs: BrowserEnv(task_class, *args, **kwargs),
        nondeterministic=task_class.nondeterministic,
        *args,
        **kwargs,
    )



import base64
import io
import logging
import numpy as np
import playwright.sync_api
import PIL.Image
import pkgutil
import re

from .constants import BROWSERGYM_ID_ATTRIBUTE as BID_ATTR

MARK_FRAMES_MAX_TRIES = 3


def _pre_extract(page: playwright.sync_api.Page):
    """
    pre-extraction routine, marks dom elements (set bid and dynamic attributes like value and checked)
    """
    js_frame_mark_elements = pkgutil.get_data(__name__, "javascript/frame_mark_elements.js").decode(
        "utf-8"
    )

    # we can't run this loop in JS due to Same-Origin Policy
    # (can't access the content of an iframe from a another one)
    def mark_frames_recursive(
        frame,
        global_iframe_position,
        iframe_offset=None,
    ):
        # get the bid of the parent frame element
        try:
            parent_bid = frame.frame_element().get_attribute(BID_ATTR)
        except:
            parent_bid = ""
        # mark all DOM elements in the frame (it will use the parent frame element's bid as a prefix)
        super_iframe_offset = frame.evaluate(
            js_frame_mark_elements,
            [
                parent_bid,
                BID_ATTR,
                global_iframe_position,
                iframe_offset,
            ],
        )

        # recursively mark all descendant frames
        for _, sub_frame in enumerate(frame.child_frames):
            if not sub_frame.is_detached():
                is_frame_hidden = sub_frame.evaluate(
                    """ () => {
    const style = window.getComputedStyle(document.documentElement);
    const is_null_size = document.documentElement.offsetWidth <= 0 || document.documentElement.offsetHeight <= 0;
    return style.display === 'none' || style.visibility === 'hidden' || is_null_size;
}"""
                )
                if not is_frame_hidden:
                    sub_iframe_position = {
                        key: sub_frame.frame_element().bounding_box()[key] for key in ["x", "y"]
                    }
                    mark_frames_recursive(sub_frame, sub_iframe_position, super_iframe_offset)

    # mark all frames recursively
    global_iframe_position = {"x": 0, "y": 0}

    mark_frames_recursive(page.main_frame, global_iframe_position)


def _post_extract(page: playwright.sync_api.Page):
    js_frame_unmark_elements = pkgutil.get_data(
        __name__, "javascript/frame_unmark_elements.js"
    ).decode("utf-8")

    # we can't run this loop in JS due to Same-Origin Policy
    # (can't access the content of an iframe from a another one)
    for frame in page.frames:
        try:
            frame.evaluate(js_frame_unmark_elements)
        except playwright.sync_api.Error as e:
            if "Frame was detached" in str(e):
                pass
            else:
                raise e


def extract_screenshot(page: playwright.sync_api.Page):
    """
    Extracts the screenshot image of a Playwright page using Chrome DevTools Protocol.

    Args:
        page: the playwright page of which to extract the screenshot.

    Returns:
        A screenshot of the page, in the form of a 3D array (height, width, rgb).

    """

    cdp = page.context.new_cdp_session(page)
    cdp_answer = cdp.send(
        "Page.captureScreenshot",
        {
            "format": "png",
        },
    )
    cdp.detach()

    # bytes of a png file
    png_base64 = cdp_answer["data"]
    png_bytes = base64.b64decode(png_base64)
    with io.BytesIO(png_bytes) as f:
        # load png as a PIL image
        img = PIL.Image.open(f)
        # convert to RGB (3 channels)
        img = img.convert(mode="RGB")
        # convert to a numpy array
        img = np.array(img)

    return img


# TODO: handle more data items if needed
__BID_EXPR = r"([-0-9]+)"
__FLOAT_EXPR = r"([+-]?(?:[0-9]*[.])?[0-9]+)"
__BOOL_EXPR = r"([01])"
# bid, bbox_left, bbox_top, center_x, center_y, bbox_right, bbox_bottom, is_in_viewport
__DATA_REGEXP = re.compile(
    __BID_EXPR
    + r"_"
    + __FLOAT_EXPR
    + r"_"
    + __FLOAT_EXPR
    + r"_"
    + __FLOAT_EXPR
    + r"_"
    + __FLOAT_EXPR
    + r"_"
    + __FLOAT_EXPR
    + r"_"
    + __FLOAT_EXPR
    + r"_"
    + __BOOL_EXPR
    + r"_"
    + r"(.*)"
)


def extract_data_items_from_aria(string):
    """
    Utility function to extract temporary data stored in the "aria-roledescription" attribute of a node
    """

    match = __DATA_REGEXP.fullmatch(string)
    if not match:
        logging.warning(
            f'Data items could not be extracted from "aria-roledescription" attribute: {string}'
        )
        return [], string

    groups = match.groups()
    data_items = groups[:-1]
    original_aria = groups[-1]
    return data_items, original_aria


def extract_dom_snapshot(
    page: playwright.sync_api.Page,
    computed_styles=[],
    include_dom_rects: bool = True,
    include_paint_order: bool = True,
    temp_data_cleanup: bool = True,
):
    """
    Extracts the DOM snapshot of a Playwright page using Chrome DevTools Protocol.

    Args:
        page: the playwright page of which to extract the screenshot.
        computed_styles: whitelist of computed styles to return.
        include_dom_rects: whether to include DOM rectangles (offsetRects, clientRects, scrollRects) in the snapshot.
        include_paint_order: whether to include paint orders in the snapshot.
        temp_data_cleanup: whether to clean up the temporary data stored in the "aria-roledescription" attribute.

    Returns:
        A document snapshot, including the full DOM tree of the root node (including iframes,
        template contents, and imported documents) in a flattened array, as well as layout
        and white-listed computed style information for the nodes. Shadow DOM in the returned
        DOM tree is flattened.

    """
    cdp = page.context.new_cdp_session(page)
    dom_snapshot = cdp.send(
        "DOMSnapshot.captureSnapshot",
        {
            "computedStyles": computed_styles,
            "includeDOMRects": include_dom_rects,
            "includePaintOrder": include_paint_order,
        },
    )
    cdp.detach()

    # if requested, remove temporary data stored in the "aria-roledescription" attribute of each node
    if temp_data_cleanup:
        try:
            target_attr_name_id = dom_snapshot["strings"].index("aria-roledescription")
        except ValueError:
            target_attr_name_id = -1
        # run the cleanup only if the "aria-roledescription" string is present
        if target_attr_name_id > -1:
            processed_string_ids = set()
            for document in dom_snapshot["documents"]:
                for node_attributes in document["nodes"]["attributes"]:
                    i = 0
                    # find the "aria-roledescription" attribute, if any
                    while i < len(node_attributes):
                        attr_name_id = node_attributes[i]
                        attr_value_id = node_attributes[i + 1]
                        if attr_name_id == target_attr_name_id:
                            attr_value = dom_snapshot["strings"][attr_value_id]
                            # remove any data stored in the "aria-roledescription" attribute
                            if attr_value_id not in processed_string_ids:
                                _, new_attr_value = extract_data_items_from_aria(attr_value)
                                dom_snapshot["strings"][
                                    attr_value_id
                                ] = new_attr_value  # update the string in the metadata
                                processed_string_ids.add(
                                    attr_value_id
                                )  # mark string as processed (in case several "aria-roledescription" attributes share the same value string)
                            # remove "aria-roledescription" attribute (name and value) if empty
                            if new_attr_value == "":
                                del node_attributes[i : i + 2]
                            # once "aria-roledescription" is found, exit the search
                            break
                        i += 2

    return dom_snapshot


def extract_all_frame_axtrees(page: playwright.sync_api.Page):
    """
    Extracts the AXTree of all frames (main document and iframes) of a Playwright page using Chrome DevTools Protocol.

    Args:
        page: the playwright page of which to extract the frame AXTrees.

    Returns:
        A dictionnary of AXTrees (as returned by Chrome DevTools Protocol) indexed by frame IDs.

    """
    cdp = page.context.new_cdp_session(page)

    # extract the frame tree
    frame_tree = cdp.send(
        "Page.getFrameTree",
        {},
    )

    # extract all frame IDs into a list
    # (breadth-first-search through the frame tree)
    frame_ids = []
    root_frame = frame_tree["frameTree"]
    frames_to_process = [root_frame]
    while frames_to_process:
        frame = frames_to_process.pop()
        frames_to_process.extend(frame.get("childFrames", []))
        # extract the frame ID
        frame_id = frame["frame"]["id"]
        frame_ids.append(frame_id)

    # extract the AXTree of each frame
    frame_axtrees = {
        frame_id: cdp.send(
            "Accessibility.getFullAXTree",
            {"frameId": frame_id},
        )
        for frame_id in frame_ids
    }

    cdp.detach()

    # extract browsergym properties (bids, coordinates, etc.) from the "roledescription" property ("aria-roledescription" attribute)
    for ax_tree in frame_axtrees.values():
        for node in ax_tree["nodes"]:
            # look for the "roledescription" property
            if "properties" in node:
                for i, prop in enumerate(node["properties"]):
                    if prop["name"] == "roledescription":
                        data_items, new_value = extract_data_items_from_aria(prop["value"]["value"])
                        prop["value"]["value"] = new_value
                        # remove the "roledescription" property if empty
                        if new_value == "":
                            del node["properties"][i]
                        # add all extracted "browsergym" properties to the AXTree
                        if data_items:
                            (
                                browsergym_id,
                                left,
                                top,
                                center_x,
                                center_y,
                                right,
                                bottom,
                                is_in_viewport,
                            ) = data_items
                            node["properties"].append(
                                {
                                    "name": "browsergym_id",
                                    "value": {
                                        "type": "string",
                                        "value": browsergym_id,
                                    },
                                }
                            )
                            node["properties"].append(
                                {
                                    "name": "browsergym_center",
                                    "value": {
                                        "type": "list",
                                        "value": (float(center_x), float(center_y)),
                                    },
                                }
                            )
                            node["properties"].append(
                                {
                                    "name": "browsergym_bounding_box",
                                    "value": {
                                        "type": "list",
                                        "value": (
                                            float(left),
                                            float(top),
                                            float(right),
                                            float(bottom),
                                        ),
                                    },
                                }
                            )
                            node["properties"].append(
                                {
                                    "name": "browsergym_is_in_viewport",
                                    "value": {
                                        "type": "boolean",
                                        "value": bool(is_in_viewport == "1"),
                                    },
                                }
                            )
    return frame_axtrees


def extract_merged_axtree(page: playwright.sync_api.Page):
    """
    Extracts the merged AXTree of a Playwright page (main document and iframes AXTrees merged) using Chrome DevTools Protocol.

    Args:
        page: the playwright page of which to extract the merged AXTree.

    Returns:
        A merged AXTree (same format as those returned by Chrome DevTools Protocol).

    """
    frame_axtrees = extract_all_frame_axtrees(page)

    cdp = page.context.new_cdp_session(page)

    # merge all AXTrees into one
    merged_axtree = {"nodes": []}
    for ax_tree in frame_axtrees.values():
        merged_axtree["nodes"].extend(ax_tree["nodes"])
        # connect each iframe node to the corresponding AXTree root node
        for node in ax_tree["nodes"]:
            if node["role"]["value"] == "Iframe":
                frame_id = cdp.send(
                    "DOM.describeNode", {"backendNodeId": node["backendDOMNodeId"]}
                )["node"]["frameId"]
                # it seems Page.getFrameTree() from CDP omits certain Frames (empty frames?)
                # if a frame is not found in the extracted AXTrees, we just ignore it
                if frame_id in frame_axtrees:
                    # root node should always be the first node in the AXTree
                    frame_root_node = frame_axtrees[frame_id]["nodes"][0]
                    assert frame_root_node["frameId"] == frame_id
                    node["childIds"].append(frame_root_node["nodeId"])
                else:
                    logging.warning(f"Extracted AXTree does not contain frameId '{frame_id}'")

    cdp.detach()

    return merged_axtree


def extract_focused_element_bid(page: playwright.sync_api.Page):
    # this JS code will dive through ShadowDOMs
    extract_focused_element_with_bid_script = """\
() => {
    // This recursive function traverses shadow DOMs
    function getActiveElement(root) {
        const active_element = root.activeElement;

        if (!active_element) {
            return null;
        }

        if (active_element.shadowRoot) {
            return getActiveElement(active_element.shadowRoot);
        } else {
            return active_element;
        }
    }
    return getActiveElement(document);
}"""
    # this playwright code will dive through iFrames
    frame = page
    focused_bid = ""
    while frame:
        focused_element = frame.evaluate_handle(
            extract_focused_element_with_bid_script, BID_ATTR
        ).as_element()
        if focused_element:
            frame = focused_element.content_frame()
            focused_bid = focused_element.get_attribute(BID_ATTR)
        else:
            frame = None

    return focused_bid
